// catdsp is a tool for configuring a /dev/dsp device
// and piping the data from this device to stdout.
// written 2001-09 by Thorsten Reinecke
// Version 1.2b, last change: 2002-11 by Thorsten Reinecke
// Copyright (C) 2001,2002 by Thorsten Reinecke (reinecke@thorstenreinecke.de)

#define CATDSP_VERSION "1.2b"

/*
    This program is distributed under the GNU GENERAL PUBLIC LICENSE (GPL)
    Version 2 (June 1991). See the "COPYING" file distributed with this
    software for more info.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
*/


#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <sys/soundcard.h>
#include <sys/ioctl.h>
#include <getopt.h>
#include <signal.h>


static int alive = 1;

/*
 Signal handler for catching Kill-Signals (SIGTERM, SIGQUIT, SIGABRT, SIGINT).
*/
static void catch_SIGTERM (int signo)
{
  alive=0; 
}


const char* const option_string = "+hvd:c:s:m:ilwrT:L:G:S:A";
static const struct option long_options[] =
{
  { "help", 0, 0, 'h' },
  { "verbose", 0, 0, 'v'},
  { "device", 1, 0, 'd' },
  { "mixer", 1, 0, 'm' },
  { "channels", 1, 0, 'c' },
  { "samplerate", 1, 0, 's' },
  { "micro", 0, 0, 'i' },
  { "line", 0, 0, 'l' },
  { "wav", 0, 0, 'w' },
  { "raw", 0, 0, 'r' },
  { "Trigger", 1, 0, 'T'},
  { "Length", 1, 0, 'L'},
  { "Gap", 1, 0, 'G'},
  { "Silence", 1, 0, 'S'},
  { "Autoadjust", 0, 0, 'A'},
  { 0, 0, 0, 0 }
}; // refer "man 3 getopt" for this struct

void usage(const char* const prg)
{
  printf("catdsp is a tool for configuring a /dev/dsp device\n");
  printf("and piping the data from this device to stdout.\n");
  printf("Version %s, (C) 2002 Thorsten Reinecke (reinecke@thorstenreinecke.de)\n",CATDSP_VERSION);
  printf("usage: \n");
  printf("%s  [-h] [-d <devicename>] [-r <sample rate>]\n",prg);
  printf("  -h, --help               this text\n");
  printf("  -d, --device <device>    defaults to /dev/dsp\n");
  printf("  -m, --mixer <device>     defaults to /dev/mixer\n");
  printf("  -c, --channels <1..n>    default: 2 (stereo)\n");
  printf("  -s, --samplerate <sample rate> defaults to 44100\n");
  printf("  -i --micro | -l --line   default: micro as input\n");
  printf("  -w, --wav   (default)    creates a wav bytestream\n");
  printf("  -r, --raw                creates a raw bytestream\n\n");
  printf("  Special functions (only 16bit stereo!)\n");
  printf("  -T, --Trigger <0-2000>   trigger record on input noise\n");
  printf("      0=off, default noise threshold to detect start: 25\n");
  printf("  -L, --Length <secs>      mininal length of track\n");
  printf("  -G, --Gap <1/10secs>     stop after a gap of silence\n");
  printf("  -S, --Silence <20-2000>  default noise threshold to detect gap: 850\n");
  printf("  -A, --Autoadjust         adjust recording level\n\n");
}

// here are the defaults:
int verbose=0;
int wav_mode = 1; // 0=raw, 1=wav
int samplerate = 44100;
int bits_per_sample = 16;
int channels = 2; // default: stereo
int blocksize; // wird spter vom Treiber erfragt
const char* device = "/dev/dsp";
const char* mixer_device = "/dev/mixer";
int dsp_fd = -1;
int input_source = SOUND_MASK_MIC; // default: micro 

void write_wav_header(int fd)
{
  char* p = strdup("RIFF    WAVEfmt                     data    ");
  int bytes_per_second=(bits_per_sample/8)*samplerate*channels;
  *(long*)&p[4]=(long)12;
  *(long*)&p[16]=(long)16;
  *(short*)&p[20]=(short)1;
  *(short*)&p[22]=(short)channels;
  *(int*)&p[24]=(int)samplerate;
  *(int*)&p[28]=(int)bytes_per_second;
  *(short*)&p[34]=(short)bits_per_sample;
  *(long*)&p[40]=(long)(3*60*60*bytes_per_second); // assume max. 3h
  write(fd,p,44);
}

int set_mixer(void)
{
  int mix_fd = open(mixer_device, O_WRONLY, 0);
  fprintf(stderr,"setting mixer to %i\n",input_source);
  int p=input_source;
  int i=ioctl(mix_fd, SOUND_MIXER_WRITE_RECSRC, &p);
  close(mix_fd);
  return (i==0);
}

typedef struct TLevel
{
  int l;
  int r;
} TLevel; 

int adjust_igain(TLevel delta)
{
  int mix_fd = open(mixer_device, O_RDWR);
  fprintf(stderr,"adjusting igain by l:%i r:%i\n",delta.l,delta.r);
  int level;
  int i=ioctl(mix_fd, SOUND_MIXER_READ_IGAIN, &level);
  level+=delta.l; level+=(int)delta.r<<8;
  i+=ioctl(mix_fd, SOUND_MIXER_WRITE_IGAIN, &level);
  close(mix_fd);
  return (i==0);
}

int open_dsp(void)
{
  dsp_fd = open(device, O_RDONLY, 0);
  if (dsp_fd == -1)
   {
     fprintf(stderr,"Cannot open %s\n",device);
     return 0;
   }
  else
   {
     int p;
     int i=ioctl(dsp_fd, SNDCTL_DSP_RESET, 0);
     p=bits_per_sample; i+=ioctl(dsp_fd, SNDCTL_DSP_SAMPLESIZE, &p);
     p=channels; i+=ioctl(dsp_fd, SNDCTL_DSP_CHANNELS, &p);
     p=samplerate; i+=ioctl(dsp_fd, SNDCTL_DSP_SPEED, &p);
     i+=ioctl(dsp_fd, SNDCTL_DSP_GETBLKSIZE, &blocksize);
     i+=ioctl(dsp_fd, SNDCTL_DSP_SYNC, 0);
     return (i==0);
   }
}


typedef struct
{
  short int l;
  short int r;
} TStereosample; 

int SilenceGapThreshold = 850;
int TriggerRecordThreshold = 25;
int autoadjust = 0;
double Minimalzeit = 60*60*3; // 3h fr die Aufnahme vorsehen
double Gap = 2.0;
double Aufnahmezeit = 0.0;
double Silence_am_Block = 0.0;

void auswerten(TStereosample* buf, int nr_samples, int autoadjust, int SilenceThreshold)
{
  const double secs = (double)nr_samples/(double)samplerate;
  Aufnahmezeit+=secs;
  if (channels !=2 || bits_per_sample!=16) return;
  
  static double Aufnahmebereitschaftszeit = 0.0;
  Aufnahmebereitschaftszeit+=secs;
  
  static TStereosample min = {r:0,l:0}, max = {r:0,l:0};
  int i;
  for (i=0; i<nr_samples; ++i)
   {
     if (buf[i].r<min.r) min.r=buf[i].r;
     if (buf[i].r>max.r) max.r=buf[i].r;
     if (buf[i].l<min.l) min.l=buf[i].l;
     if (buf[i].l>max.l) max.l=buf[i].l;
   }

  // nun "Amplituden" ermitteln:
  int dr = (int)max.r - (int)min.r;
  int dl = (int)max.l - (int)min.l;
  int dmax = dr<dl ? dl : dr; // hier steht die maximale Differenz der Samples "Amplitude"

  if (dmax>SilenceThreshold) // hier mu irgendein vernnftiger Treshold eingesetzt werden
   Silence_am_Block=0.0; // keine Silence
  else
   Silence_am_Block+=secs;
  

  static double next_Auswertung = 0.1;
  if (Aufnahmebereitschaftszeit < next_Auswertung) return;
  min.r=min.l=max.r=max.l=0; // wieder resetten
  if (verbose||Silence_am_Block) fprintf(stderr,"dl: %i  dr: %i  max: %i  Silence: %g\n", dl, dr, dmax, Silence_am_Block);
  next_Auswertung+=0.1;

  
  if (autoadjust)
   { 
     static TLevel Sperre = { l:0, r:0 };
     // zunchst den Fall zu hoher Lautstrke...
     if (dr>40000 || dl>40000) // bersteuert?
      {
        TLevel level = {l:0,r:0};
        if (dl>40000) { level.l=-1; Sperre.l=50; }
        if (dr>40000) { level.r=-1; Sperre.r=50; }
        adjust_igain(level);
      }

     const double interval = 0.5;
     const int Sollpegel = 33000;
     const int delta = 100000;
     // Heuristik:
     // durchschnittliche Maximalpegel innerhalb von "interval" ermitteln,
     // pro 2000er Abweichung vom Sollpegel ein Tick in die Gegenrichtung.
     // Adjustment nur durchfhren, wenn zum vorherigen Adjustment(wunsch)
     // auf mindestens einem Kanal keine Differenz besteht.
     static double next_check = 2.5;
     static int z = 0;
     static TLevel pegel = { l:0, r:0 };
     pegel.l+=dl; pegel.r+=dr; ++z;
     
     if (Aufnahmebereitschaftszeit>=next_check && z>1)
      {
        --Sperre.l; --Sperre.r;
        pegel.l/=z; pegel.r/=z;
        if (verbose) fprintf(stderr,"Pegelstnde %i / %i\n",pegel.l,pegel.r);

        static TLevel counter = { l:0, r:0 };

        // wenn kein Aufwrts- oder Abwrtstrend, dann darf
        // neu ausgeregelt werden, dies pro Kanal!
        
        TLevel adl = { l:0, r:0 };
        counter.l+=pegel.l-Sollpegel;
        counter.r+=pegel.r-Sollpegel;
        
        if (counter.l < -delta) { counter.l=0; if (Sperre.l<=0) adl.l=+1; }
        else if (counter.l > delta) { adl.l=-1; counter.l=0; }
        
        if (counter.r < -delta) { counter.r=0; if (Sperre.r<=0) adl.r=+1; }
        else if (counter.r > delta) { adl.r=-1; counter.r=0; }
        
        if (verbose) fprintf(stderr,"Counterstnde %i / %i\n",counter.l,counter.r);
        
        if (adl.l || adl.r) adjust_igain(adl);

        z=pegel.l=pegel.r=0;
        next_check=Aufnahmebereitschaftszeit+interval;
      }
   }

  //fprintf(stderr,"Aufnahmezeit: %g  SZ: %g\n", Aufnahmezeit, Silence_am_Block);
}

int main(int argc, char* argv[])
{
  int option_index;
  int o;
  while ( (o=getopt_long(argc,argv,option_string,long_options,&option_index)) != -1)
   {
     switch (o)
      {
        case 'h': usage(argv[0]); exit(0); break;
        case 'v': verbose=1; break;
        case 'd': device=strdup(optarg); break;
        case 'm': mixer_device=strdup(optarg); break;
        case 'c': channels=atoi(optarg); break;
        case 's': samplerate=atoi(optarg); break; 
        case 'i': input_source=SOUND_MASK_MIC; break;
        case 'l': input_source=SOUND_MASK_LINE; break;
        case 'w': wav_mode=1; break;
        case 'r': wav_mode=0; break;
        case 'T': TriggerRecordThreshold=atoi(optarg); break;
        case 'L': Minimalzeit=atoi(optarg); break;
        case 'G': Gap=0.1*atoi(optarg); break;
        case 'S': SilenceGapThreshold=atoi(optarg); break;
        case 'A': autoadjust=1; break;
       default: usage(argv[0]); exit(1);
      }
   }
  // now optind points to the first argv-element that is not an option (refer man getopt)

  if (optind<argc)
   {
     fprintf(stderr," - too many parameters.\n");
     usage(argv[0]); exit(1); 
   }

  if (!set_mixer()) exit(2);
  if (!open_dsp()) exit(2);
  
  signal(SIGTERM, catch_SIGTERM); // handle termination signal
  signal(SIGQUIT, catch_SIGTERM); // handle quit-signal from keyboard
  signal(SIGABRT, catch_SIGTERM); // handle abort signal
  signal(SIGINT, catch_SIGTERM); // handle interrupt signal from keyboard
  
  fprintf(stderr,"Blocksize: %i\n",blocksize);
  if (wav_mode) write_wav_header(1);
  
  {
    const int size = blocksize;
    const int samples_per_buf = size/4; // works only for 16 bit stereo for now
    char buf[size];
    
    if (TriggerRecordThreshold)
     {
       int count;
       do
        {
          count=read(dsp_fd,buf,size);
          if (count<=0) break;
          if (count!=blocksize) fprintf(stderr,"count: %i\n",count);
          auswerten((TStereosample*)buf,samples_per_buf,0,TriggerRecordThreshold);
        } while (alive && Silence_am_Block>0.0);
       write(1,buf,count);
     }
     
    fprintf(stderr,"start recording...\n"); 
    fprintf(stderr,"Wartezeit: %g\n", Aufnahmezeit);
    Aufnahmezeit=0.0;
    
    while (alive && (Aufnahmezeit<=Minimalzeit || Silence_am_Block<Gap))
     {
       int count=read(dsp_fd,buf,size);
       if (count<=0) break;
       if (count!=blocksize) fprintf(stderr,"count: %i\n",count);
       auswerten((TStereosample*)buf,samples_per_buf,autoadjust,SilenceGapThreshold);
       write(1,buf,count);
     }
    fprintf(stderr,"Aufnahmezeit: %g  Silencetime: %g\n", Aufnahmezeit, Silence_am_Block);
  }
 close(dsp_fd);
 return 0;
}
